#include "InterlockedCallbackQueue.h"
#include <assert.h>
#include <time.h>

RCSTATUS InterlockedCallbackQueue::EnqueueCallback( 
		ICallback* pCb )
{
	assert( pCb );
	if ( ! pCb ) return RCSTATUS_INVALID_PARAMETER;
	
	if ( RCSTATUS_SUCCESS != m_state )
	{
		return m_state;
	}
	
	//
	// acquire lock
	//
	int res = pthread_mutex_lock( &m_queueLock );
	if ( EOK != res )
	{
		return MAKE_RCSTATUS( 
			RCTSTATUS_FACILITY_CBQ, res );
	}
	
	//
	// enqueue
	//
	m_queue.push( pCb );
	
	//
	// release lock
	//
	res = pthread_mutex_unlock( &m_queueLock );
	if ( EOK != res )
	{
		return MAKE_RCSTATUS( 
			RCTSTATUS_FACILITY_CBQ, res );
	}
	
	//
	// notify that queue is non-empty now
	//
	if ( 0 != sem_post( &m_sem ) )
	{
		return RCSTATUS_FROM_ERRNO();
	}
	
	//
	// evth fine
	//
	return RCSTATUS_SUCCESS;
}

RCSTATUS InterlockedCallbackQueue::DequeueCallback( 
		DWORD dwTimeout, 
		ICallback **ppCb )
{
	assert( ppCb );
	if ( ! ppCb ) return RCSTATUS_INVALID_PARAMETER;
	*ppCb = NULL;
	
	if ( RCSTATUS_SUCCESS != m_state )
	{
		return m_state;
	}
	
	int res = 0;
	
	if ( dwTimeout > 0 )
	{
		// sem_wait wants an absolute date, so calculate one
		timespec ts;
		if ( -1 == clock_gettime( CLOCK_REALTIME, &ts ) )
		{
			return RCSTATUS_FROM_ERRNO();
		}
		ts.tv_sec += dwTimeout;
		
		// 
		// wait for semaphore/for queue to contain an object
		//
		res = sem_timedwait( &m_sem, &ts );
		if ( 0 != res )
		{
			if ( ETIMEDOUT == errno )
			{
				//
				// wait timed out, just return
				//
				return RCSTATUS_SUCCESS;
			}
			else 
			{
				return RCSTATUS_FROM_ERRNO();
			}
		}
	}
	else
	{
		//
		// check if queue contains an object, else return
		//
		res = sem_trywait( &m_sem );
		if ( 0 != res )
		{
			//
			// sem locked -> queue empty
			//
			return RCSTATUS_SUCCESS;
		}
	}
	
	//
	// now go ahead and dequeue object
	//
	
	//
	// acquire lock
	//
	res = pthread_mutex_lock( &m_queueLock );
	if ( EOK != res )
	{
		return MAKE_RCSTATUS( 
			RCTSTATUS_FACILITY_CBQ, res );
	}
	
	//
	// dequeue
	//
	if ( m_queue.empty() )
	{
		//
		// weird - sem was signaled, though the queue
		// is empty. Ok to continue though.
		//
		assert( !"CBQ: semaphore was signaled though "
			"the queue is empty" );
		*ppCb = NULL;
	}
	else
	{
		*ppCb = m_queue.front();
		m_queue.pop();
	}
	//
	// release lock
	//
	res = pthread_mutex_unlock( &m_queueLock );
	if ( EOK != res )
	{
		return MAKE_RCSTATUS( RCSTATUS_CBQ_NOT_INITIALIZED, res );
	}
	
	return RCSTATUS_SUCCESS;
}

//RCSTATUS InterlockedCallbackQueue::UnblockAll()
//{
//	if ( RCSTATUS_SUCCESS != m_state )
//	{
//		return m_state;
//	}
//	return RCSTATUS_SUCCESS;
//}
