#ifndef __INTERLOCKEDCALLBACKQUEUE_H__
#define __INTERLOCKEDCALLBACKQUEUE_H__

#include <pthread.h>
#include <semaphore.h>
#include <queue>
#include <errno.h>
#include "ICallback.h"

#define RCTSTATUS_FACILITY_CBQ	 		1
#define RCSTATUS_CBQ_NOT_INITIALIZED MAKE_RCSTATUS( RCTSTATUS_FACILITY_CBQ, 1 )
	
//
// queue holding callback objects encapsulating callbacks
// to be executed asynchronously
//
// This class is threadsafe
//
class InterlockedCallbackQueue
{
private:
	//
	// queue holding callback objects
	//
	std::queue< ICallback* > m_queue;
	
	//
	// mutex for interlocking m_queue
	//
	pthread_mutex_t m_queueLock;
	
	//
	// semaphore for blocking on empty queue
	//
	sem_t m_sem;
	
	RCSTATUS m_state;

public:
	InterlockedCallbackQueue()
	{
		//
		// initialize semaphore (unshared, initially locked)
		//
		if ( 0 != sem_init( &m_sem, 0, 0 ) )
		{
			m_state = RCSTATUS_FROM_ERRNO();
		}
		
		//
		// initialize mutex
		//
		int res = pthread_mutex_init( &m_queueLock, NULL );
		if ( EOK != res )
		{
			m_state = MAKE_RCSTATUS(
				RCTSTATUS_FACILITY_CBQ, res );
		}
		
		m_state = RCSTATUS_SUCCESS;
	}
	
	~InterlockedCallbackQueue()
	{
		sem_close( &m_sem );
		pthread_mutex_destroy( &m_queueLock );
	}
	
	//
	// Add a callback object to the end of the queue
	//
	RCSTATUS EnqueueCallback( 
		ICallback* pCb );
	
	// 
	// retrieve callback from begin of the queue.	
	// If the queue is empty, the behaviour is specified
	// by dwTimeout:
	//   dwTimeout = 0 
	//      -> return with *ppCmd = NULL
	//   TIMEOUT_INFINITE > dwTimeout > 0 
	//      -> block until a callback is available or the
	//         timeout elapses. In the latter case, *pCmd 
	//         set to NULL
	//   dwTimout = TIMEOUT_INFINITE
	//      -> ditto, but *pCmd is never st to NULL unless
	//         a status other than RCSTATUS_SUCCESS is returned
	//
	// see also: UnblockAll()
	//
	RCSTATUS DequeueCallback( 
		DWORD dwTimeout, 
		ICallback **ppCb );

	
	//
	// all threads blocking in DequeueCallback are unblocked.
	// 
	// RCSTATUS UnblockAll();
};


#endif // __INTERLOCKEDCALLBACKQUEUE_H__
