#include "SerialPort.h"
#include <atomic.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include "Std.h"

#define MAX_WAIT 1000
#define MAX_ITERATIONS_WITHOUT_WAITING 100

void SerialPort::DumpDeviceRegisters()
{
	int data;

	fprintf( stderr, "Modem Status Register\n" );
	data = in8( m_pPort + 6 );

	fprintf( stderr, "  Carrier detect:               %d\n", ( data & 128 ) > 0 );
	fprintf( stderr, "  Ring indicator:               %d\n", ( data & 64  ) > 0 );
	fprintf( stderr, "  Data set ready:               %d\n", ( data & 32  ) > 0 );
	fprintf( stderr, "  Clear to send:                %d\n", ( data & 16  ) > 0 );
	fprintf( stderr, "  Delta data carrier detect:    %d\n", ( data & 8   ) > 0 );
	fprintf( stderr, "  Trailing edge ring indicator: %d\n", ( data & 4   ) > 0 );
	fprintf( stderr, "  Delta data set ready:         %d\n", ( data & 2   ) > 0 );
	fprintf( stderr, "  Delta clear to send:          %d\n", ( data & 1   ) > 0 );
	
	fprintf( stderr, "Line Status Register\n" );
	data = in8( m_pPort + 5 );

	fprintf( stderr, "  Error in received FIFO:       %d\n", ( data & 128 ) > 0 );
	fprintf( stderr, "  Empty data holding registers: %d\n", ( data & 64 )  > 0 );
	fprintf( stderr, "  Empty tranm. holding register:%d\n", ( data & 32 )  > 0 );
	fprintf( stderr, "  Break interrupt:              %d\n", ( data & 16 )  > 0 );
	fprintf( stderr, "  Framing error:                %d\n", ( data & 8 )   > 0 );
	fprintf( stderr, "  Parity error:                 %d\n", ( data & 4 )   > 0 );
	fprintf( stderr, "  Overrun error:                %d\n", ( data & 2 )   > 0 );
	fprintf( stderr, "  Data ready:                   %d\n", ( data & 1 )   > 0 );
	
	fprintf( stderr, "Interrupt Identification Register\n" );
	data = in8( m_pPort + 5 );
	
	switch ( ( data >> 1 ) & 0x3 )
	{
		case 0:	
			fprintf( stderr, "  Modem Status ITR\n" ); break;
		case 1:	
			fprintf( stderr, "  Transm. Holding Reg. empty ITR\n" ); break;
		case 2:	
			fprintf( stderr, "  Data available ITR\n" ); break;
		case 3:	
			fprintf( stderr, "  Line Status ITR\n" ); break;
	}
	fprintf( stderr, "\n\n" );
	fflush( stderr );
}

/*----------------------------------------------------------------------
 *
 * Private - Open/Close port
 *
 *--------------------------------------------------------------------*/

RCSTATUS SerialPort::SetupPort()
{
	//
	// gain thread privity - neccessary for accessing device
	//
	if ( -1 ==  ThreadCtl( _NTO_TCTL_IO, NULL ) )
	{
		return RCSTATUS_FROM_ERRNO();
	}

	//
	// open port
	//
	uintptr_t pPort = mmap_device_io( PORT_LENGTH, m_dwIoAddr );
	if ( MAP_DEVICE_FAILED == pPort )
	{
		return RCSTATUS_FROM_ERRNO();
	}

	m_pPort = pPort;

	//
	// configure port
	//
	out8( pPort + 3, 0x80 );	// set DLAB 
	out8( pPort + 0, 0x30 ); 	// set baud rate to 2400
	out8( pPort + 1, 0x00 ); 	// set baud rate to 2400
	out8( pPort + 3, 0x07 );	// set LCR (8bit, 2stopbits)
	out8( pPort + 2, 0xc7 );	// set FCR (8bit + clear)
	out8( pPort + 4, 0x0b );	// enable DTR, RTS, OUT2
	out8( pPort + 1, 13 );		// enable all interrupts

	return RCSTATUS_SUCCESS;
}

RCSTATUS SerialPort::ClosePort()
{
	if ( m_pPort )
	{
		int ret = munmap_device_io( m_pPort, PORT_LENGTH );
		m_pPort = NULL;

		if ( -1 == ret )
		{
			return RCSTATUS_FROM_ERRNO();
		}
	}

	return RCSTATUS_SUCCESS;
}

/*----------------------------------------------------------------------
 *
 * Private - Thread spawning
 *
 *--------------------------------------------------------------------*/

RCSTATUS SerialPort::StartHandlerThread()
{
	//
	// just spawn thread
	//
	int res = pthread_create( 
		&m_handlerThread, 
		NULL, 
		InterruptHandlerThreadProc, 
		this );
	if ( EOK != res )
	{
		return MAKE_RCSTATUS( 
			RCTSTATUS_FACILITY_PORT, res );
	}

	return RCSTATUS_SUCCESS;
}

RCSTATUS SerialPort::StopHandlerThread()
{
	//
	// set flag to indicate the handler thread to stop pumping
	//
	atomic_set( &m_fStopPump, 1 );

	//
	// wait for thread
	//
	int res = pthread_join( m_handlerThread, NULL );
	if ( EOK != res )
	{
		return MAKE_RCSTATUS( 
			RCSTATUS_FACILITY_ERRNO, res );
	}

	return RCSTATUS_SUCCESS;
}


/*----------------------------------------------------------------------
 *
 * I/O handling thread (pump)
 *
 *--------------------------------------------------------------------*/

void* SerialPort::InterruptHandlerThreadProc( void* pSerialPort )
{
	SerialPort* pMe = reinterpret_cast< SerialPort* >( pSerialPort );
	assert( pMe );
	
	//
	// start pumping
	//
	pMe->Pump();
	
	return NULL;
}

void SerialPort::Pump()
{
	//
	// 1. set up port
	//
	RCSTATUS rcs = SetupPort();
	if ( RCSTATUS_SUCCESS == rcs )
	{		
		//
		// 2. enter pump
		//
		rcs = PumpLoop();
		if ( RCSTATUS_SUCCESS != rcs )
		{
			m_handler.InternalError( rcs );
		}

		//
		// close port
		//
		rcs = ClosePort();
		if ( RCSTATUS_SUCCESS != rcs )
		{
			//
			// tell handler and stop
			//
			m_handler.InternalError( rcs );
		}
	}
	else
	{
		//
		// setting up port failed - tell handler and stop
		//
		m_handler.InternalError( rcs );
	}
}

RCSTATUS SerialPort::PumpLoop()
{
	//
	// pump until m_fStopPump is set
	//
	
	//
	// dispatch while the Modem Status Register (MSR) and 
	// Line Status Register (LSR) tell that there is sth to do
	//
	int msr = 0;
	int lsr = 0;
	
	DWORD dwMillisToWait = 0;
	DWORD dwIterationsWithoutWaiting = 0;
	for ( ; ; )
	{
		// 
		// is it time to stop?
		//
		if ( 1 == atomic_clr_value( &m_fStopPump, 1 ) )
		{
			//
			// stop pump
			//
			return RCSTATUS_SUCCESS;
		}

		msr = in8( m_pPort + 6 );
		lsr = in8( m_pPort + 5 );
		//
		// decide what to do
		//
		if ( lsr & 1 )
		{
			//
			// data available
			//
			if ( 0 == dwMillisToWait )
				dwIterationsWithoutWaiting++;
			if ( MAX_ITERATIONS_WITHOUT_WAITING == dwIterationsWithoutWaiting )
			{
				fprintf( stderr, "PANIC: Hardware is going mad!\n" );
				DumpDeviceRegisters();
				return RCSTATUS_PORT_PANIC;;
			}
			
			dwMillisToWait = 0;	// reset
			
			
			RCSTATUS rcs = m_handler.Receive( m_pPort );
			if ( RCSTATUS_SUCCESS != rcs )
			{
				// tear down
				return rcs;
			}
		}
		else if ( 
			( msr & 16 ) && ( lsr & 32 ) ) // Clear To Send
		{
			//
			// ready to send
			//
			if ( 0 == dwMillisToWait )
				dwIterationsWithoutWaiting++;
			if ( MAX_ITERATIONS_WITHOUT_WAITING == dwIterationsWithoutWaiting )
			{
				fprintf( stderr, "PANIC: Hardware is going mad!\n" );
				DumpDeviceRegisters();
				return RCSTATUS_PORT_PANIC;;
			}
			
			dwMillisToWait = 0;	// reset
			
			RCSTATUS rcs = m_handler.Send( m_pPort );
			if ( RCSTATUS_SUCCESS != rcs )
			{
				// tear down
				return rcs;
			}
		}
		else if ( lsr & 8 )	// Framing Error
		{
			dwMillisToWait = 0;	// reset
			
			return m_handler.DeviceFramingError();
		}
		else if ( lsr & 4 )	// Parity Error
		{
			dwMillisToWait = 0;	// reset
			
			return m_handler.DeviceFramingError();
		}
		else if ( lsr & 1 )	// Overrun Error
		{
			dwMillisToWait = 0;	// reset
			
			return m_handler.DeviceFramingError();
		}
		else
		{
			//
			// wait a little - each time a little longer
			//
			m_handler.Wait( dwMillisToWait );
			nap( dwMillisToWait );
			if ( dwMillisToWait  < MAX_WAIT )
				dwMillisToWait++;

			dwIterationsWithoutWaiting = 0;
		}
	}
}

/*----------------------------------------------------------------------
 *
 * Publics
 *
 *--------------------------------------------------------------------*/

RCSTATUS SerialPort::Open()
{
	//
	// spawn handler thread to begin feeding the handler
	//
	RCSTATUS rcs = StartHandlerThread();
	if ( RCSTATUS_SUCCESS != rcs )
	{
		return rcs;
	}

	rcs = m_handler.Start();
	if ( RCSTATUS_SUCCESS != rcs )
	{
		return rcs;
	}
	
	return RCSTATUS_SUCCESS;
}

RCSTATUS SerialPort::Close()
{
	//
	// stop handler thread 
	//
	RCSTATUS rcs = StopHandlerThread();
	if ( RCSTATUS_SUCCESS != rcs )
	{
		return rcs;
	}

	rcs = m_handler.Stop();
	if ( RCSTATUS_SUCCESS != rcs )
	{
		return rcs;
	}

	return RCSTATUS_SUCCESS;
}
