#include <stdio.h>
#include "SerialPort.h"
#include "SerialPortHandler.h"
#include "Bus.h"

RCSTATUS SerialPortHandler::DismissCurrentCommand(
		IN BYTE *pResponseData,
		IN DWORD dwResponseDataLen)
{
	RCSTATUS rcs = m_pCurrentCommand->Complete( 
		pResponseData, dwResponseDataLen );

	delete m_pCurrentCommand;
	m_pCurrentCommand = NULL;
	m_dwNextCommandByteIndex = 0;
		
	return rcs;
}

RCSTATUS SerialPortHandler::NextCommandByte( 
		INOUT BYTE *pb,
		INOUT BOOL *pfEndOfCommand )
{
	assert( pb );
	if ( ! pb ) return RCSTATUS_INVALID_PARAMETER;
	*pb = NULL;

	if ( ! m_pCurrentCommand )
	{
		//
		// no command available, fetch one
		// the dispenser is expected to ALWAYS return a command 
		// on success
		//
		RCSTATUS rcs = m_dispenser.NextCommand( &m_pCurrentCommand );
		if ( RCSTATUS_SUCCESS != rcs )
		{
			return rcs;
		}
		
		assert( m_pCurrentCommand );
		if ( ! m_pCurrentCommand ) return RCSTATUS_UNEXPECTED;
	}

	//
	// How long is this command (#bytes)?
	//
	DWORD dwCmdLen = 0;
	RCSTATUS rcs = m_pCurrentCommand->GetData( 
		NULL, &dwCmdLen, 0 );
	if ( RCSTATUS_SUCCESS != rcs )
	{
		return rcs;
	}
	
	//
	// try to fetch byte from current command
	//
	DWORD dwBufLen = 1;
	rcs = m_pCurrentCommand->GetData( 
		pb, &dwBufLen, m_dwNextCommandByteIndex );
	
	if ( RCSTATUS_SUCCESS == rcs )
	{
		// 
		// do not read the same byte again
		//
		m_dwNextCommandByteIndex++;
	}

	//
	// check this was the last byte
	//
	*pfEndOfCommand = ( m_dwNextCommandByteIndex >= dwCmdLen );
	return rcs;
}

RCSTATUS SerialPortHandler::Start()
{
	return RCSTATUS_SUCCESS;
}

RCSTATUS SerialPortHandler::Stop()
{
	printf( "Bytes sent:     %u\n"
			"Bytes received: %u\n"
			"Waited:         %u\n" , 
			( unsigned int ) m_dwBytesSent,
			( unsigned int ) m_dwBytesReceived,
			( unsigned int ) m_dwTimeSpentWaiting );
	return RCSTATUS_SUCCESS;
}


RCSTATUS SerialPortHandler::Send( uintptr_t pPort )
{
	//DBG_PRINT( "SerialPortHandler::Send (enter)\n" );
	if ( Sending != m_state )
	{
		return RCSTATUS_SUCCESS;
	}
	//DBG_PRINT( "SerialPortHandler::Send\n" );
		
	BYTE bToSend = 0;
	BOOL fEndOfCommand = FALSE;
	RCSTATUS rcs = NextCommandByte( &bToSend, &fEndOfCommand );
	if ( RCSTATUS_SUCCESS != rcs ) 
	{
		return rcs;
	}

	//
	// output 
	//
	//printf( "Sending 0x%x\n", bToSend );
	out8( pPort, bToSend );
	m_dwBytesSent++;

	if ( ! fEndOfCommand )
	{
		//
		// command not fully sent yet
		//
	}
	else
	{
		// 
		// command fully sent -> check what to do next
		//
		DWORD dwRespLen = 0;
		rcs = m_pCurrentCommand->GetExpectedResponseLength( &dwRespLen );
		if ( RCSTATUS_SUCCESS != rcs ) 
		{
			return rcs;
		}
	
		if ( dwRespLen > 0 )
		{
			//
			// this command expects some response -> 
			// Write() will dismiss it
			//
			// enter receiving state
			//
			m_state = Receiving;
			rcs = RCSTATUS_SUCCESS;
		}
		else
		{
			//
			// write-only-command -> dismiss it now
			//
			rcs = DismissCurrentCommand( NULL, 0 );
		}
	}
	return rcs;
}

RCSTATUS SerialPortHandler::Receive( uintptr_t pPort )
{
	if ( ! m_pCurrentCommand )
	{
		//
		// we are receiving data though we are not supposed to - huh?
		//
		return RCSTATUS_UNEXPECTED;
	}
	
	//DBG_PRINT( "SerialPortHandler::Receive\n" );
	//
	// How many bytes does the command expect?
	//
	DWORD dwRespLen = 0;
	RCSTATUS rcs = m_pCurrentCommand->GetExpectedResponseLength( &dwRespLen );
	if ( RCSTATUS_SUCCESS != rcs ) 
	{
		return rcs;
	}

	if ( m_dwNextRcvBufferIndex == 
		sizeof( m_rgbRcvBuffer ) / sizeof( BYTE ) )
	{
		//
		// buffer is about to overflow
		//
		return RCSTATUS_BUF_OVERFLOW;
	}

	//
	// read into buffer
	//
	m_rgbRcvBuffer[ m_dwNextRcvBufferIndex++ ] = in8( pPort );
	m_dwBytesReceived++;

	// printf( "received: %x\n", m_rgbRcvBuffer[ m_dwNextRcvBufferIndex - 1 ] );
	//
	// more expected?
	//
	BOOL fEndOfResponse = 
		( m_dwNextRcvBufferIndex >= dwRespLen );
	
	if ( fEndOfResponse )
	{
		//DBG_PRINT( "SerialPortHandler::Receive - done\n" );		
		//
		// leave receiving state
		//
		m_state = Sending;
		
		//
		// dismiss command
		//
		rcs = DismissCurrentCommand( m_rgbRcvBuffer, dwRespLen );
	
		m_dwNextRcvBufferIndex = 0;
		
		return rcs;
	}
	else
	{
		//
		// more data expected
		//
		return RCSTATUS_SUCCESS;
	}
}

RCSTATUS SerialPortHandler::DeviceFramingError()
{
	fprintf( stderr,  "ERROR: Device framing error\n" );
	return RCSTATUS_FRAMING_ERROR;
}

RCSTATUS SerialPortHandler::DeviceParityError()
{
	fprintf( stderr,  "ERROR: Parity error\n" );
	return RCSTATUS_PARITY_ERROR;
}

RCSTATUS SerialPortHandler::DeviceOverflowError()
{
	fprintf( stderr,  "ERROR: Overflow error\n" );
	return RCSTATUS_OVERFLOW_ERROR;
}

RCSTATUS SerialPortHandler::InternalError( RCSTATUS rcs )
{
	fprintf( stderr,  "ERROR: Internal error RCSTATUS 0x%x\n", rcs  );
	return RCSTATUS_SUCCESS;
}

void SerialPortHandler::Wait( DWORD dwMillis )
{
	m_dwTimeSpentWaiting += dwMillis;
}
